#include "includes.h"

#include "moteur.h"
#include "utils.h"
#include "types.h"
#include "inputs.h"

// Concernant le code provenant de https://lodev.org/cgtutor/raycasting.html, c'est a dire move(), draw_walls()
/*
Copyright (c) 2004-2021, Lode Vandevenne

All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
// moteur.c :
// ici se trouvent tout ce qui concerne les graphismes, logique, mouvement et collisions
//
//

image_t *tex_index[tindex_size];

int8_t on_solid_ground(ShooterMap *ShooterLevel, Vector3d pos) {
  // If pos is int and tile under is either solid or z is 0
  if(fix(ffloor(pos.z)) == pos.z){
    if(pos.z < fix(1)) return 1;
    if(has_collision(
      ShooterLevel->wall[ffloor(pos.z - 0x0001)][ffloor(pos.x)][ffloor(pos.y)]
      ) > 0){
      return 1;
    }
  }
  return 0;
}
int8_t bumps_head(ShooterMap *ShooterLevel, Vector3d pos) {
  pos.z += player_height;
  if(fix(ffloor(pos.z)) == pos.z){
    if(pos.z > fix(3)) return 1;
    if(has_collision(
      ShooterLevel->wall[ffloor(pos.z + 0x0001)][ffloor(pos.x)][ffloor(pos.y)]
      ) > 0){
      return 1;
    }
  }
  return 0;
}

void move(Game *game, uint8_t keys) {
  Player *player = game->player;
  ShooterMap *shooterLevel = game->shooterLevel;

  extern int frame_time;
  fixed_t sFrameTime = fix(frame_time)/1000; //Frame time in seconds
  fixed_t moveSpeed = fmul(sFrameTime, 20*player->speed_mult); //frame_time * fix(carrés/seconde là carrés/seconde = 5
  fixed_t rotSpeed = sFrameTime*2; //frame_time * fix(radians/seconde là radians/seconde = 2
  fixed_t c_rotSpeed = fix(cos(f2float(rotSpeed)));
  fixed_t s_rotSpeed = fix(sin(f2float(rotSpeed)));

  fixed_t oldDirX;
  fixed_t oldPlaneX;
  Vector3d fpos;
  Vector3d temp;
  uint8_t jumped;

  if(keys & 0b1) {
    player->velocity.x = fmul(player->dir.x, moveSpeed);
    player->velocity.y = fmul(player->dir.y, moveSpeed);
  }
  if(keys & 0b10) {
    player->velocity.x = -fmul(player->dir.x, moveSpeed);
    player->velocity.y = -fmul(player->dir.y, moveSpeed);
  }
  //rotate to the right
  if(keys & 0b1000) {
    //both camera direction and camera plane must be rotated
    oldDirX = player->dir.x;
    player->dir.x = (fmul(player->dir.x, c_rotSpeed)+1) - (fmul(player->dir.y, -s_rotSpeed)+1);
    player->dir.y = (fmul(oldDirX, -s_rotSpeed)+1) + (fmul(player->dir.y, c_rotSpeed)+1);
    oldPlaneX = player->plane.x;
    player->plane.x = (fmul(player->plane.x, c_rotSpeed)+1) - (fmul(player->plane.y, -s_rotSpeed)+1);
    player->plane.y = (fmul(oldPlaneX, -s_rotSpeed)+1) + (fmul(player->plane.y, c_rotSpeed)+1);
  }
  //rotate to the left
  if(keys & 0b100) {
    //both camera direction and camera plane must be rotated
    oldDirX = player->dir.x;
    player->dir.x = (fmul(player->dir.x, c_rotSpeed)-1) - (fmul(player->dir.y, s_rotSpeed)-1);
    player->dir.y = (fmul(oldDirX, s_rotSpeed)+1) + (fmul(player->dir.y, c_rotSpeed)+1);
    oldPlaneX = player->plane.x;
    player->plane.x = (fmul(player->plane.x, c_rotSpeed)-1) - (fmul(player->plane.y, s_rotSpeed) - 1);
    player->plane.y = (fmul(oldPlaneX, s_rotSpeed)+1) + (fmul(player->plane.y, c_rotSpeed) + 1);
  }
  //Jump
  if(keys & 0b10000) {
    if(on_solid_ground(shooterLevel, player->pos) > 0){
      player->velocity.z = moveSpeed;
      jumped = 1;
      dprint(1,150,C_BLACK,"jumped");
    }
    dprint(1,140,C_BLACK,"tried to jump");
  }

  temp.x = ffloor(player->pos.x + fmul(player->velocity.x, sFrameTime));
  temp.y = ffloor(player->pos.y + fmul(player->velocity.y, sFrameTime));
  temp.z = ffloor(player->pos.z + fmul(player->velocity.z, sFrameTime));
  fpos.x = ffloor(player->pos.x);
  fpos.y = ffloor(player->pos.y);
  fpos.z = ffloor(player->pos.z);

  player->velocity.x -= fmul(player->velocity.x, fmul(0x5000, fsq(player->velocity.x)));
  player->velocity.y -= fmul(player->velocity.y, fmul(0x5000, fsq(player->velocity.x)));

  if(shooterLevel->wall[fpos.z][temp.x][fpos.y] == 0) player->pos.x = player->pos.x + fmul(player->velocity.x, sFrameTime);
  else player->velocity.x = 0;
  if(shooterLevel->wall[fpos.z][fpos.x][temp.y] == 0) player->pos.y = player->pos.y + fmul(player->velocity.y, sFrameTime);
  else player->velocity.y = 0;

  /*if((on_solid_ground(shooterLevel, (Vector3d){player->pos.x, player->pos.y, temp.z}) > 0
      || bumps_head(shooterLevel, (Vector3d){player->pos.x, player->pos.y, temp.z}) < 1)&&
      jumped < 1) {
    player->velocity.z = 0;
  }
  else player->velocity.z -= 2*fmul(G, sFrameTime);*/ //Gravité

  if (player->dir.x > 0xFFFF) player->dir.x = 0xFFFF;
  if (player->dir.y > 0xFFFF) player->dir.y = 0xFFFF;
  if (player->dir.x < -0xFFFF) player->dir.x = -0xFFFF;
  if (player->dir.y < -0xFFFF) player->dir.y = -0xFFFF;
}

void draw_background(int back_id){ //a refaire

}

void logic(){
  //logique (logique)
}

void draw_f(image_t *vram){ //a refaire

}

void draw_walls(Game *game, image_t *vram){
  Player *player = game->player;
  ShooterMap *shooterLevel = game->shooterLevel;

  fixed_t cameraX;
  fixed_t rayDirX;
  fixed_t rayDirY;
  fixed_t sideDistX;//length of ray from current position to next x or y-side
  fixed_t sideDistY;
  fixed_t deltaDistX;
  fixed_t deltaDistY;
  fixed_t perpWallDist;
  fixed_t texSize;
  int x;
  int i;
  int mapX;
  int mapY;
  int stepX; //what direction to step in x or y-direction (either +1 or -1)
  int stepY;
  int side; //was a NS or a EW wall hit?
  int lineHeight;
  int texX;
  int texSample;
  int texSampleY;

  int v_offset = 0; //(int)(sin(f2int(posX + posY)) * 5); //a raffiner un peu
  fixed_t h_offset = 0; //fix(sin(f2int(posX - posY)) * 0.01);

  uint8_t *dbuffer[viewport_h];

  struct image_linear_map temp;

  image_t *tex_stripe;

  extern image_t *tex_index[tindex_size];

  extern int raycast_time;
  extern int draw_time;

  prof_t rayscat = prof_make();
  prof_t img_drw = prof_make();

  for(x = 0; x < viewport_w; x+=2) {
    prof_enter(rayscat);

    debug_func();

    //calculate ray position and direction
    cameraX = fdiv(fix(x*2), fix(viewport_w)) - 0xFFFF + h_offset; //x-coordinate in camera space

    rayDirX = player->dir.x + fmul(player->plane.x, cameraX);
    rayDirY = player->dir.y + fmul(player->plane.y, cameraX);


    //which box of the map we're in
    mapX = f2int(player->pos.x);
    mapY = f2int(player->pos.y);

    // length of ray from one x or y-side to next x or y-side
    // these are derived as:
    // deltaDistX = sqrt(1 + (rayDirY * rayDirY) / (rayDirX * rayDirX))
    // deltaDistY = sqrt(1 + (rayDirX * rayDirX) / (rayDirY * rayDirY))
    // which can be simplified to abs(|rayDir| / rayDirX) and abs(|rayDir| / rayDirY)
    // where |rayDir| is the length of the vector (rayDirX, rayDirY). Its length,
    // unlike (dirX, dirY) is not 1, however this does not matter, only the
    // ratio between deltaDistX and deltaDistY matters, due to the way the DDA
    // stepping further below works. So the values can be computed as below.
    // Division through zero is prevented, even though technically that's not
    // needed in C++ with IEEE 754 floating point values.
    // Fcalva : removed the 0 div prevention

    deltaDistX = abs(fdiv(0xFFFF, rayDirX));
    deltaDistY = abs(fdiv(0xFFFF, rayDirY));

    //calculate step and initial sideDist
    if (rayDirX < 0) {
    stepX = -1;
    sideDistX = fmul(player->pos.x - fix(mapX), deltaDistX);
    }
    else {
      stepX = 1;
      sideDistX = fmul(fix(mapX + 1) - player->pos.x, deltaDistX);
    }

    if (rayDirY < 0) {
      stepY = -1;
      sideDistY = fmul(player->pos.y - fix(mapY), deltaDistY);
    }
    else {
      stepY = 1;
      sideDistY = fmul( fix(mapY + 1) - player->pos.y, deltaDistY);
    }

    for(i=0; i<viewport_h; i++){
      dbuffer[i] = 0x7FFFFFFF;
    }

    //perform DDA
    while(1) {
      //Check if the ray is out of range/bounds
      if (sideDistX >= max_dist || sideDistY >= max_dist || mapX < 0 || mapY < 0 || mapX >= map_w || mapY >= map_h) {
        break;
      }
      //Otherwise check if ray has hit walls, and draw them
      else if(shooterLevel->wall[0][mapX][mapY] != 255){
        prof_leave(rayscat);
        prof_enter(img_drw);

        //Calculate distance projected on camera direction. This is the shortest distance from the point where the wall is
        //hit to the camera plane. Euclidean to center camera point would give fisheye effect!
        //This can be computed as (mapX - posX + (1 - stepX) / 2) / rayDirX for side == 0, or same formula with Y
        //for size == 1, but can be simplified to the code below thanks to how sideDist and deltaDist are computed:
        //because they were left scaled to |rayDir|. sideDist is the entire length of the ray above after the multiple
        //steps, but we subtract deltaDist once because one step more into the wall was taken above.

        if (side == 0) perpWallDist = (sideDistX - deltaDistX);
        else perpWallDist = (sideDistY - deltaDistY);

        //texturing calculation
        //calculate value of wallX
        fixed_t wallX; //where exactly the wall was hit
        if (side == 0) wallX = player->pos.y + fmul(perpWallDist, rayDirY);
        else  wallX = player->pos.x + fmul(perpWallDist, rayDirX);
        wallX -= fix(ffloor(wallX));

          //x coordinate on the texture
        texX = fround(wallX * tsize);

        if(side == 0 && rayDirX > 0) texX = tsize - texX - 1;
        if(side == 1 && rayDirY < 0) texX = tsize - texX - 1;

        perpWallDist = perpWallDist < 1 ? 1 : perpWallDist;

        lineHeight = fround(fdiv(fix(viewport_h), perpWallDist)); //Taille en px de la ligne
        if (lineHeight < 1) lineHeight = 1;
        if (lineHeight > viewport_h-1) lineHeight = viewport_h-1;

        int glinePos = viewport_h/2 - lineHeight/2;

        fixed_t texSize = fix(lineHeight) / tsize; //taille proportionelle de la ligne a la tex

        for(i = 0; i < map_d; i++) {

          if(!has_collision(shooterLevel->wall[i][mapX][mapY])) continue;

          int zOffset = player->pos.z - fix(i);

          // The reasoning here is to offset the wall slice vertically from the center of the screen,
          // depending on the offset between the wall and the player
          int linePos = glinePos + fround(zOffset*lineHeight);

          if(linePos >= viewport_h) continue;

          if(linePos+lineHeight < 1) continue;

          //if(dbuffer[linePos]<perpWallDist && dbuffer[linePos+lineHeight]<perpWallDist) continue;

          uint8_t tex = 0; //ShooterLevel->wall[mapX][mapY][i];

          if (linePos < 1) {
            texSampleY = texSize*abs(linePos);
            texSample = tsize - texSampleY;
            linePos = 1;
          }
          else {
            texSampleY = 0;
            texSample = 64;
          }

          color_t color;

          switch(i){
            case 0 : {color = C_PINK; break;}
            case 1 : {color = C_BLACK; break;}
            case 2 : {color = C_BSKY; break;}
          }

          //char ymin = 0;
          //char ymax = viewport_h;
          //int midline = (linePos*2+lineHeight)/2;

          for(int j=linePos+2; j<linePos+lineHeight-2; j+=4){
            if(dbuffer[j] < perpWallDist>>14){
              //if(j <= midline) ymin = j;
              //if(j >= midline) ymax = j;
              continue;
            }
            dbuffer[j] = perpWallDist >> 14;
            dbuffer[j+1] = perpWallDist >> 14;
            dbuffer[j+2] = perpWallDist >> 14;
            dbuffer[j+3] = perpWallDist >> 14;
            drect(x,j,x+1,j+3,perpWallDist>>16);
          }

          /*image_sub(tex_index[tex], texX, texSampleY, 1, texSample, tex_stripe);

          image_scale(tex_stripe, 0xFFFF, texSize, &temp);

          image_linear(tex_stripe, image_at(vram, x, linePos), &temp);*/

          //prof_leave(img_drw);
          //prof_enter(rayscat);
        }
      }
      //jump to next map square, either in x-direction, or in y-direction
      if (sideDistX < sideDistY) {
        sideDistX += deltaDistX;
        mapX += stepX;
        side = 0;
      }
      else {
        sideDistY += deltaDistY;
        mapY += stepY;
        side = 1;
      }
    }
    //prof_leave(rayscat);
  }

  //raycast_time = (int)prof_time(rayscat)/1000;
}
