#include "includes.h"

#ifdef SDL2

#define window_w (screen_w*upscale)
#define window_h (screen_h*upscale)
#define SDL_viewport_w (viewport_w*upscale)
#define SDL_viewport_h (viewport_h*upscale)

SDL_Info SDL_inf;

void sdl_image_init(){
  extern SDL_Info SDL_inf;

  uint32_t init_flags = SDL_INIT_TIMER | SDL_INIT_AUDIO | SDL_INIT_VIDEO;

  SDL_Init(init_flags);

  uint32_t window_flags = SDL_WINDOW_INPUT_GRABBED;

  SDL_inf.window = SDL_CreateWindow("RaycasterGame", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED,
                                       window_w, window_h, window_flags);

  uint32_t renderer_flags = SDL_RENDERER_TARGETTEXTURE;

  SDL_inf.renderer = SDL_CreateRenderer(SDL_inf.window, -1, renderer_flags);

  TTF_Init();

  SDL_inf.font = TTF_OpenFont("./assets-cg/StarmapTruetype.ttf", 8);
}

void sdl_image_quit(){
  extern SDL_Info SDL_inf;

  TTF_CloseFont(SDL_inf.font);

  TTF_Quit();

  IMG_Quit();

  SDL_DestroyRenderer(SDL_inf.renderer);

  SDL_DestroyWindow(SDL_inf.window);

  SDL_Quit();
}

image_t sdl_tex_load(char *filename){
  extern SDL_Info SDL_inf;

  image_t tex;
  int tformat, taccess, w, h;

  tex.tex = IMG_LoadTexture(SDL_inf.renderer, filename);

  tex.subrect = malloc(sizeof(SDL_Rect));

  SDL_QueryTexture(tex.tex, &tformat, &taccess, &w, &h);

  tex.subrect->x = 0;
  tex.subrect->y = 0;
  tex.subrect->w = w;
  tex.subrect->h = h;

  tex.is_vram = 0;

  return tex;
}

image_t *image_alloc(int w, int h, enum img_type image_type){
  extern SDL_Info SDL_inf;

  image_t *tex;
  tex = malloc(sizeof(image_t));

  tex->subrect = malloc(sizeof(SDL_Rect));

  tex->subrect->x = 0;
  tex->subrect->y = 0;
  tex->subrect->w = w;
  tex->subrect->h = h;

  tex->tex = SDL_CreateTexture(SDL_inf.renderer, SDL_PIXELFORMAT_RGBA8888, SDL_TEXTUREACCESS_TARGET,
                                 w, h);

  return tex;
}

void image_free(image_t *img){

  free(img->subrect);

  SDL_DestroyTexture(img->tex);

  free(img);
}

int prof_init(void){
  return 0;
}

void prof_quit(void){
   return;
}

image_t *image_create_vram(void){
  image_t *tex;
  tex = NULL;

  return tex;
}

void dclear(SDL_Color color){
  extern SDL_Info SDL_inf;

  SDL_SetRenderDrawColor(SDL_inf.renderer, color.r, color.g, color.b, color.a);
  SDL_RenderClear(SDL_inf.renderer);
  SDL_SetRenderDrawColor(SDL_inf.renderer, 255, 255, 255, 255);
}

void dupdate(void){
  extern SDL_Info SDL_inf;

  SDL_RenderPresent(SDL_inf.renderer);
}

static SDL_Rect dSubrect;

void dimage(int x, int y, image_t *image){
  extern SDL_Info SDL_inf;

  dSubrect.x = x;
  dSubrect.y = y;
  dSubrect.w = image->subrect->w;
  dSubrect.h = image->subrect->h;

  SDL_RenderCopy(SDL_inf.renderer, image->tex, image->subrect, &dSubrect);
}

static image_t image_sub_default;

image_t *image_sub(image_t *src, int x, int y, int w, int h, image_t *dest){
  if(!dest) dest = &image_sub_default;

  if(w < 0) w = src->subrect->w - x;
  if(h < 0) h = src->subrect->h - y;


  dest->subrect->x = x;
  dest->subrect->y = y;
  dest->subrect->w = w;
  dest->subrect->h = h;

  dest->tex = src->tex;

  return dest;
}

void image_scale(image_t const *src, int gamma_x, int gamma_y, struct image_linear_map *map){
  map->h_scale = f2float(gamma_x) * upscale;
  map->v_scale = f2float(gamma_y) * upscale;
}

void image_linear(image_t const *src, image_t *dst, struct image_linear_map *map){
  extern SDL_Info SDL_inf;

  dSubrect.x = 0;
  dSubrect.y = 0;
  dSubrect.w = round(src->subrect->w * map->h_scale);
  dSubrect.h = round(src->subrect->h * map->v_scale);

  SDL_RenderCopy(SDL_inf.renderer, src->tex, src->subrect, &dSubrect);
}

void dprint(int x, int y, SDL_Color fg, char const *format, ...){
  extern SDL_Info SDL_inf;

  SDL_Surface *tsurface;
  SDL_Texture *ttex;
  uint32_t tformat;
  int w, h, taccess;
  char formatted[256];

  snprintf(formatted, sizeof(char)*256, format);

  tsurface = TTF_RenderText_Blended(SDL_inf.font, formatted, fg);

  ttex = SDL_CreateTextureFromSurface(SDL_inf.renderer, tsurface);

  SDL_QueryTexture(ttex, &tformat, &taccess, &w, &h);

  dSubrect.x = x;
  dSubrect.y = y;
  dSubrect.w = w;
  dSubrect.h = h;

  SDL_RenderCopy(SDL_inf.renderer, ttex, NULL, &dSubrect);

  SDL_FreeSurface(tsurface);

  SDL_DestroyTexture(ttex);
}

void gint_dvline(int y1, int y2, int x, SDL_Color color){
  extern SDL_Info SDL_inf;

  SDL_SetRenderDrawColor(SDL_inf.renderer, color.r, color.g, color.b, color.a);

  SDL_RenderDrawLine(SDL_inf.renderer, x, y1, x, y2);

  SDL_SetRenderDrawColor(SDL_inf.renderer, 255, 255, 255, 255);
}

#endif
