#include "includes.h"

#ifndef utils__h
#define utils__h

#include "types.h"

#define sq(x) (x*x)
#define fsq(x) (fmul(x,x))

//Provient de
//https://stackoverflow.com/questions/31117497/fastest-integer-square-root-in-the-least-amount-of-instructions
//, njuffa

static const uint8_t clz_tab[32] = {
    31, 22, 30, 21, 18, 10, 29,  2, 20, 17, 15, 13, 9,  6, 28, 1,
    23, 19, 11,  3, 16, 14,  7, 24, 12,  4,  8, 25, 5, 26, 27, 0};

static inline uint8_t clz (uint32_t a)
{
    a |= a >> 16;
    a |= a >> 8;
    a |= a >> 4;
    a |= a >> 2;
    a |= a >> 1;
    return clz_tab [0x07c4acdd * a >> 27];
}

/* 16 x 16 -> 32 bit unsigned multiplication; should be single instruction */
static inline uint32_t umul16w (uint16_t a, uint16_t b)
{
    return (uint32_t)a * b;
}

/* Reza Hashemian, "Square Rooting Algorithms for Integer and Floating-Point
   Numbers", IEEE Transactions on Computers, Vol. 39, No. 8, Aug. 1990, p. 1025
*/
static uint16_t isqrt (uint32_t x)
{
    volatile uint16_t y, z, lsb, mpo, mmo, lz, t;

    if (x == 0) return x; // early out, code below can't handle zero

    lz = clz (x);         // #leading zeros, 32-lz = #bits of argument
    lsb = lz & 1;
    mpo = 17 - (lz >> 1); // m+1, result has roughly half the #bits of argument
    mmo = mpo - 2;        // m-1
    t = 1 << mmo;         // power of two for two's complement of initial guess
    y = t - (x >> (mpo - lsb)); // initial guess for sqrt
    t = t + t;            // power of two for two's complement of result
    z = y;

    y = (umul16w (y, y) >> mpo) + z;
    y = (umul16w (y, y) >> mpo) + z;
    if (x >= 0x40400) {
        y = (umul16w (y, y) >> mpo) + z;
        y = (umul16w (y, y) >> mpo) + z;
        if (x >= 0x1002000) {
            y = (umul16w (y, y) >> mpo) + z;
            y = (umul16w (y, y) >> mpo) + z;
        }
    }

    y = t - y; // raw result is 2's complement of iterated solution
    y = y - umul16w (lsb, (umul16w (y, 19195) >> 16)); // mult. by sqrt(0.5)

    if ((int32_t)(x - umul16w (y, y)) < 0) y--; // iteration may overestimate
    if ((int32_t)(x - umul16w (y, y)) < 0) y--; // result, adjust downward if
    if ((int32_t)(x - umul16w (y, y)) < 0) y--; // necessary

    return y; // (int)sqrt(x)
}

static inline fixed_t fsqrt(fixed_t x){
  return isqrt(x) << 8;
}

static inline fixed_t fix2Ddotp(Vector2d a, Vector2d b){
  return fmul(a.x, b.x) + fmul(a.y, b.y);
}

static inline fixed_t fix2DNorm(Vector2d a) {
  return fsqrt(fix2Ddotp(a, a));
}

static inline Vector2d fix2DNorml(Vector2d a) {
  Vector2d b;
  fixed_t ilen = 1/fix2DNorm(a);
  b.x = fmul(a.x ,ilen);
  b.y = fmul(a.y, ilen);
  return b;
}

int cmpfunc(const void *, const void *);

fixed_t raycast(ShooterMap*, fixed_t, fixed_t, fixed_t, fixed_t, fixed_t, char);

#ifdef FXCG50

typedef uint16_t color_t;

void *bsearch (const void *key, const void *base0, size_t nmemb, size_t size,
         int (*compar)(const void *, const void *));

static inline color_t rgb565to8888(uint16_t a){
  return a;
}

#endif
#ifdef SDL2

typedef SDL_Color color_t;

static color_t rgb565to8888(uint16_t a){
  uint8_t r,g,b;

  r = (a & 0b1111000000000) >> 11;
  g = (a & 0b0000111110000) >> 5;
  b =  a & 0b0000000001111;

  SDL_Color c = {r, g, b, 255};

  return c;
}

#endif

//Function to trigger GDB breakpoints
void debug_func();

#endif /* sprites.h */
